import { when } from 'jest-when';
import { pick } from 'lodash';
import { Test, TestingModule } from '@nestjs/testing';
import { getRepositoryToken } from '@nestjs/typeorm';
import { Repository } from 'typeorm';
import {
  mockCaCertificate, mockCaCertificateCertificateEncrypted, mockCaCertificateCertificatePlain, mockCaCertificateEntity,
  mockCaCertificateId, mockEncryptionService,
  mockRepository,
  MockType,
} from 'src/__mocks__';
import { LocalCaCertificateRepository } from 'src/modules/certificate/repositories/local.ca-certificate.repository';
import { CaCertificateEntity } from 'src/modules/certificate/entities/ca-certificate.entity';
import { EncryptionService } from 'src/modules/encryption/encryption.service';
import { BadRequestException, NotFoundException } from '@nestjs/common';
import ERROR_MESSAGES from 'src/constants/error-messages';

describe('LocalCaCertificateRepository', () => {
  let service: LocalCaCertificateRepository;
  let encryptionService: MockType<EncryptionService>;
  let repository: MockType<Repository<CaCertificateEntity>>;

  beforeEach(async () => {
    jest.clearAllMocks();

    const module: TestingModule = await Test.createTestingModule({
      providers: [
        LocalCaCertificateRepository,
        {
          provide: getRepositoryToken(CaCertificateEntity),
          useFactory: mockRepository,
        },
        {
          provide: EncryptionService,
          useFactory: mockEncryptionService,
        },
      ],
    }).compile();

    repository = await module.get(getRepositoryToken(CaCertificateEntity));
    encryptionService = await module.get(EncryptionService);
    service = await module.get(LocalCaCertificateRepository);

    repository.findOneBy.mockResolvedValue(mockCaCertificateEntity);
    repository.createQueryBuilder().getMany.mockResolvedValue([
      pick(mockCaCertificateEntity, 'id', 'name'),
      pick(mockCaCertificateEntity, 'id', 'name'),
    ]);
    repository.save.mockResolvedValue(mockCaCertificateEntity);
    repository.create.mockReturnValue(mockCaCertificate); // not entity since it happens before encryption

    when(encryptionService.decrypt).calledWith(mockCaCertificateCertificateEncrypted, jasmine.anything())
      .mockResolvedValue(mockCaCertificateCertificatePlain);
    when(encryptionService.encrypt).calledWith(mockCaCertificateCertificatePlain)
      .mockResolvedValue({
        data: mockCaCertificateCertificateEncrypted,
        encryption: mockCaCertificateEntity.encryption,
      });
  });

  describe('get', () => {
    it('should return ca certificate model', async () => {
      const result = await service.get(mockCaCertificateId);

      expect(result).toEqual(mockCaCertificate);
    });
  });

  describe('list', () => {
    it('should return ca certificates list', async () => {
      const result = await service.list();

      expect(result).toEqual([
        pick(mockCaCertificate, 'id', 'name'),
        pick(mockCaCertificate, 'id', 'name'),
      ]);
    });
  });

  describe('create', () => {
    it('should create ca certificate', async () => {
      repository.findOneBy.mockResolvedValueOnce(null);

      const result = await service.create(mockCaCertificate);

      expect(result).toEqual(mockCaCertificate);
    });

    it('should throw an error when ca certificate with such name already exists', async () => {
      try {
        await service.create(mockCaCertificate);
        fail();
      } catch (e) {
        // todo: why not ConflictException?
        expect(e).toBeInstanceOf(BadRequestException);
        expect(e.message).toEqual(ERROR_MESSAGES.CA_CERT_EXIST);
        expect(repository.save).not.toHaveBeenCalled();
      }
    });
  });

  describe('delete', () => {
    it('should delete ca certificate', async () => {
      const result = await service.delete(mockCaCertificate.id);

      expect(result).toEqual(undefined);
    });

    it('should throw an error when trying to delete non-existing ca certificate', async () => {
      repository.findOneBy.mockResolvedValueOnce(null);

      try {
        await service.delete(mockCaCertificate.id);
        fail();
      } catch (e) {
        expect(e).toBeInstanceOf(NotFoundException);
        // todo: why such message?
        expect(e.message).toEqual('Not Found');
        expect(repository.delete).not.toHaveBeenCalled();
      }
    });
  });
});
