import React from 'react'
import { cloneDeep } from 'lodash'
import { BuildType } from 'uiSrc/constants/env'
import { localStorageService } from 'uiSrc/services'
import { setFeaturesToHighlight } from 'uiSrc/slices/app/features-highlighting'
import { getNotifications } from 'uiSrc/slices/app/notifications'
import { render, mockedStore, cleanup, MOCKED_HIGHLIGHTING_FEATURES } from 'uiSrc/utils/test-utils'

import {
  getUserConfigSettings,
  setSettingsPopupState,
  userSettingsSelector,
} from 'uiSrc/slices/user/user-settings'
import { appServerInfoSelector, getServerInfo } from 'uiSrc/slices/app/info'
import { processCliClient } from 'uiSrc/slices/cli/cli-settings'
import { getRedisCommands } from 'uiSrc/slices/app/redis-commands'
import Config from './Config'

let store: typeof mockedStore
beforeEach(() => {
  cleanup()
  store = cloneDeep(mockedStore)
  store.clearActions()
})

jest.mock('uiSrc/slices/user/user-settings', () => ({
  ...jest.requireActual('uiSrc/slices/user/user-settings'),
  userSettingsSelector: jest.fn().mockReturnValue({
    config: {
      agreements: {},
    },
    spec: {
      agreements: {},
    },
  }),
}))

jest.mock('uiSrc/slices/app/info', () => ({
  ...jest.requireActual('uiSrc/slices/app/info'),
  appServerInfoSelector: jest.fn()
}))

jest.mock('uiSrc/services', () => ({
  ...jest.requireActual('uiSrc/services'),
  localStorageService: {
    set: jest.fn(),
    get: jest.fn(),
  },
}))

describe('Config', () => {
  it('should render', () => {
    render(<Config />)
    const afterRenderActions = [
      getServerInfo(),
      processCliClient(),
      getRedisCommands(),
      getNotifications(),
      getUserConfigSettings()
    ]
    expect(store.getActions()).toEqual([...afterRenderActions])
  })

  it('should call the list of actions', () => {
    const userSettingsSelectorMock = jest.fn().mockReturnValue({
      config: {
        agreements: {},
      },
      spec: {
        agreements: {
          eula: {
            defaultValue: false,
            required: true,
            editable: false,
            since: '1.0.0',
            title: 'EULA: RedisInsight License Terms',
            label: 'Label',
          },
        },
      },
    })
    userSettingsSelector.mockImplementation(userSettingsSelectorMock)
    render(<Config />)
    const afterRenderActions = [
      getServerInfo(),
      processCliClient(),
      getRedisCommands(),
      getNotifications(),
      getUserConfigSettings(),
      setSettingsPopupState(true),
    ]
    expect(store.getActions()).toEqual([...afterRenderActions])
  })

  it('should call updateHighlightingFeatures for new user with empty features', () => {
    const userSettingsSelectorMock = jest.fn().mockReturnValue({
      config: {
        agreements: null,
      }
    })
    const appServerInfoSelectorMock = jest.fn().mockReturnValue({
      buildType: BuildType.Electron,
      appVersion: '2.0.0'
    })
    userSettingsSelector.mockImplementation(userSettingsSelectorMock)
    appServerInfoSelector.mockImplementation(appServerInfoSelectorMock)

    render(<Config />)

    expect(store.getActions())
      .toEqual(expect.arrayContaining([setFeaturesToHighlight({ version: '2.0.0', features: [] })]))
  })

  it('should call updateHighlightingFeatures for existing user with proper data', () => {
    const userSettingsSelectorMock = jest.fn().mockReturnValue({
      config: {
        agreements: {},
      }
    })
    const appServerInfoSelectorMock = jest.fn().mockReturnValue({
      buildType: BuildType.Electron,
      appVersion: '2.0.0'
    })
    userSettingsSelector.mockImplementation(userSettingsSelectorMock)
    appServerInfoSelector.mockImplementation(appServerInfoSelectorMock)

    render(<Config />)

    expect(store.getActions())
      .toEqual(expect.arrayContaining([setFeaturesToHighlight({ version: '2.0.0', features: MOCKED_HIGHLIGHTING_FEATURES })]))
  })

  it('should call updateHighlightingFeatures for existing user with proper data with features from LS', () => {
    localStorageService.get = jest.fn().mockReturnValue({ version: '2.0.0', features: ['importDatabases'] })
    const userSettingsSelectorMock = jest.fn().mockReturnValue({
      config: {
        agreements: {},
      }
    })
    const appServerInfoSelectorMock = jest.fn().mockReturnValue({
      buildType: BuildType.Electron,
      appVersion: '2.0.0'
    })
    userSettingsSelector.mockImplementation(userSettingsSelectorMock)
    appServerInfoSelector.mockImplementation(appServerInfoSelectorMock)

    render(<Config />)

    expect(store.getActions())
      .toEqual(expect.arrayContaining([setFeaturesToHighlight({ version: '2.0.0', features: ['importDatabases'] })]))
  })

  it('should call updateHighlightingFeatures for existing user with proper data with features from LS for different version', () => {
    localStorageService.get = jest.fn().mockReturnValue({ version: '2.0.0', features: ['importDatabases'] })
    const userSettingsSelectorMock = jest.fn().mockReturnValue({
      config: {
        agreements: {},
      }
    })
    const appServerInfoSelectorMock = jest.fn().mockReturnValue({
      buildType: BuildType.Electron,
      appVersion: '2.0.12'
    })
    userSettingsSelector.mockImplementation(userSettingsSelectorMock)
    appServerInfoSelector.mockImplementation(appServerInfoSelectorMock)

    render(<Config />)

    expect(store.getActions())
      .toEqual(expect.arrayContaining([setFeaturesToHighlight({ version: '2.0.12', features: MOCKED_HIGHLIGHTING_FEATURES })]))
  })
})
